# Standard imports
from pandas import DataFrame, read_sql_query


class SQLAlchemy():

    def __init__(self, engine: str, host: str, port: str, database: str, username: str, password: str) -> None:
        """
            Creates and initializes a PostgreSQL engine instance that connects to the database

            Parameters:
                engine (str): The name of the database engine (postgres, trino, etc.)
                host (str): Host IP address
                port (str): Port number
                database (str): Name of the database
                username (str): Username for authentication/privileges
                password (str): Password for authentication
        """
        self.engine = engine
        self.username = username
        self.password = password
        self.host = host
        self.port = port
        self.database = database


    def _select(self, query: str, index_col: str = None) -> DataFrame:
        """ Executes the given query and returns the results as a DataFrame """
        return read_sql_query(sql = query, con = self.engine, index_col = index_col)
    

    def _execute(self, query: str) -> None:
        """ Executes the given query """
        self.engine.execute(query)


    def _insert(self, df: DataFrame, name: str, if_exists: str = "append", index: bool = False,index_label: str = None) -> None:
        """
            Inserts the given DataFrame into the database

            Parameters:
                df (DataFrame): The data to be inserted
                name (str): The name of the table to insert the dataframe into
                if_exists (str): Whether to append to the existing table if it exists or create a new one
                index (bool): Whether to drop the index
                index_label (str): The name of the index column
        """
        df.to_sql(name, con = self.engine, if_exists = if_exists, index = index, index_label = index_label)


# Standard imports
from time import time
from typing import Callable
from datetime import datetime


# A Meta-Decorator (A decorator that can be used to decorate other decorators) is used on top of decorates to enable them to be parametrized
def parametrized(dec):
    def layer(*args, **kwargs):
        def repl(f):
            return dec(f, *args, **kwargs)
        return repl
    return layer


# Prints status and the time taken to execute the function
@parametrized
def status(func: Callable, name: str):
    """
        A decorator that logs the function's execution status onto a file

        parameters:
            name: A name that the function will be logged under
    """

    def now() -> datetime: return datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    def run(*args, **kwargs):
        try:
            print(f"[INFO - {now()}] '{name}' started..")
            start_time = time()
            func(*args, **kwargs)
            print(f"[SUCCESS - {now()}] '{name}' executed successfully. (Took {round(time() - start_time, 2)} seconds)")
        except Exception as ex:            
            print(f"[ERROR - {now()}] '{name}' failed to execute. (details: {ex})")

    return run



@parametrized
def restart_on_crash(func: Callable, name: str, notify_crash: bool = True, retries: int = None) -> None:
    """
        A decorator that restarts a function forever if it crashes for any reason

        Parameters:
            func: The function to be restarted.
            name: The name of the function for logging purposes.
            notify_crash: A boolean that indicates whether to notify the user when the function crashes.
            retries: The number of times to retry the function before giving up. If None, the function will be retried forever.
    """

    def now() -> datetime: return datetime.now().strftime('%Y-%m-%d %H:%M:%S')

    def run(*args, **kwargs):
        _retries = 0
        while True:
            try:
                result = func(*args, **kwargs)                
                return None if result is None else result   # Return the return value of the function
            except Exception as ex:
                if notify_crash:
                    print(f"[WARNING - {now()}] '{name}' crashed! restarting... (details: {ex})")
                _retries += 1   # Increment the number of retries

            # Abort the function if the number of retries is greater than the maximum number of retries
            if retries is not None and _retries >= retries:
                print(f"[ERROR - {now()}] '{name}' crashed too many timee! aborting... (max tries: {retries})")
                break

    return run


from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name = "toolkit4life",                                      # This is the name of the package
    version = "0.0.87",                                         # The initial release version
    author = "Keivan Ipchi Hagh",                               # Full name of the author
    url = "https://github.com/keivanipchihagh/toolkit4life",    # URL to the github repository
    description = "Faster deployment is what we want!",
    long_description = long_description,                        # Long description read from the the readme file
    long_description_content_type = "text/markdown",
    packages = find_namespace_packages(include = ["toolkit4life", "toolkit4life.*"]),   # List of all python modules to be installed
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],                                                          # Information to filter the project on PyPi website
    python_requires = '>=3.6',                                  # Minimum version requirement of the package
    py_modules = ["toolkit4life"],                              # Name of the python package
    # package_dir = {'':'toolkit4life'},                        # Directory of the source code of the package
    install_requires = required                                 # Install other dependencies if any
)